﻿import { Component, Output, EventEmitter, OnInit, Input } from '@angular/core';
import { ConatinerNodeModel, NodeLinkModel, ShapeNodeDataModel, NodeAttributeModel, DiagramModel, NodeHelper, NodeSizeEnum, NodeConfiguration } from '../index';

import Diagram = MindFusion.Diagramming.Diagram;
import HandlesStyle = MindFusion.Diagramming.HandlesStyle;
import Rect = MindFusion.Drawing.Rect;
import Events = MindFusion.Diagramming.Events;
import Point = MindFusion.Drawing.Point;

declare var $: any;

@Component({
    selector: 'app-diagram',
    templateUrl: './diagram.component.html'
})
export class DiagramComponent implements OnInit {

    @Output('dragstartevent') dragstartevent = new EventEmitter<any>();
    @Output('dragenterevent') dragenterevent = new EventEmitter<any>();
    @Output('dragoverevent') dragoverevent = new EventEmitter<any>();
    @Output('dropevent') dropevent = new EventEmitter<any>();
    @Output() nodeClick = new EventEmitter<any>();
    @Output() dataSourceChange = new EventEmitter();
    @Input() get dataSource() { return this.diagramModel; }
    set dataSource(value) {
        // create a Diagram component that wraps the 'diagram' canvas
        if (value !== undefined && value !== null) {
            this.diagramModel = value;
            this.diagram = Diagram.create(document.querySelector('#diagram'));
            this.diagram.setBehavior(MindFusion.Diagramming.Behavior.DoNothing);
            this.diagram.setAutoResize(MindFusion.Diagramming.AutoResize.AllDirections);
            this.diagram.setEnabled(false);
            this.prepareTreeGraph(this.diagramModel.containerNodes, this.diagramModel.links, this.diagramModel.attributes);
        }
    }

    private diagram: any;
    private incomingLinks = 0;
    private outgoingLinks = 0;
    private diagramModel: DiagramModel;
    constructor() { }

    ngOnInit() {
    }

    onDiagramDragStart(event) {
        this.dragstartevent.emit(event);
    }

    onDiagramDragOver(event) {
        this.dragoverevent.emit(event);
    }

    onDiagramDrop(event) {
        this.dropevent.emit(event);
    }

    onDiagramDragEnter(event) {
        this.dragenterevent.emit(event);
    }

    enableDrag() {
        this.diagram.setEnabled(false);
    }

    disableDrag() {
        this.diagram.setEnabled(true);
    }

    onDiagramMouseDown(event) {
        const node = this.getNodeUnderCursor(event);
        if (node) {
            const localPoint = this.getLocalPoint(event);
            const manipulator = node.hitTestManipulators(localPoint);
            this.nodeClick.emit(node.data);
            if (manipulator !== null) {
                this.diagram.setEnabled(true);
            } else {
                this.diagram.setEnabled(false);
            }
            
        } else {
            this.diagram.setEnabled(true);
        }
        
    }

    onDiagramMouseUp(event) {
        this.diagram.setEnabled(true);
    }

    getNodeUnderCursor(event) {
        const localPoint = this.getLocalPoint(event);
        return this.diagram.getNodeAt(localPoint, true, true);
    }

    getLocalPoint(event) {
        // Get the relative coordinates
        const offset = $('#diagram').offset();
        const x = (event.pageX - offset.left) + $(window).scrollLeft();
        const y = (event.pageY - offset.top) + $(window).scrollTop();

        return this.diagram.clientToDoc(new Point(x, y));
    }

    getDiagram() {
        return this.diagram;
    }
    fitItems() {
        
        let rect = MindFusion.Drawing.Rect.empty;
        this.diagram.nodes.forEach((node) => {
            if (rect == MindFusion.Drawing.Rect.empty)
                rect = node.bounds;
            else
                rect = rect.union(node.bounds);
        });

        if (rect)
            this.diagram.setBounds(new MindFusion.Drawing.Rect(0, 0, rect.right() + 10, rect.bottom() + 10));
    }
    private attachEvents(nodeAttribute: NodeAttributeModel) {
        this.diagram.addEventListener(Events.containerFolded, (sender, args) => {
            const node = args.getNode();
            const r = node.getBounds();
            r.height = nodeAttribute.height;
            r.width = nodeAttribute.width;
            node.setBounds(r, true, true);
            this.setOffsetFoldedChildNode(node, node.unfoldedSize.height - node.captionHeight);
            node.updateLinks();
        });
        this.diagram.addEventListener(Events.containerUnfolded, (sender, args) => {
            const node = args.getNode();
            const r = node.getBounds();
            r.height = node.unfoldedSize.height;
            r.width = node.unfoldedSize.width;
            node.setBounds(r, true, true);
            this.setOffsetUnFoldedChildNode(node, node.unfoldedSize.height - node.captionHeight);
            node.updateLinks();
        });
    }

    private prepareTreeGraph(containerNodes, nodeLinks, attributes) {

        if (containerNodes !== undefined) {
            const nodes: { [id: number]: any; } = {};
            if (containerNodes.length > 0) {
                const nodeAttribute: NodeAttributeModel = attributes;
                NodeConfiguration.SetSize(nodeAttribute.type);
                nodeAttribute.height = NodeConfiguration.height;
                nodeAttribute.width = NodeConfiguration.width;
                nodeAttribute.collapsibleButtonSize = NodeConfiguration.collapsibleButtonSize;
                this.attachEvents(nodeAttribute);

                let initialLocation = 20;
                containerNodes.forEach(node => {

                    let count = 1;
                    const attribute = new NodeAttributeModel();
                    attribute.inherit(nodeAttribute);
                    attribute.headingText = node.nodeTitle;
                    attribute.headingSupportText = '';
                    if (nodeAttribute.type !== NodeSizeEnum.SMALL) {
                        this.countIncomingOutgoingNodes(nodeLinks, node.nodeID);
                        attribute.incomingLinks = this.incomingLinks;
                        attribute.outgoingLinks = this.outgoingLinks;
                    }
                    const height =  node.shapeNodes.length > 1 ? 10 * node.shapeNodes.length + attribute.height : attribute.height;
                    const containerNode = new NodeHelper(this.diagram.getFactory(), attribute, height);
                    containerNode.createContainerNode(initialLocation + 20, initialLocation + 20, node.vo);

                    initialLocation += 20;
                    if (nodeAttribute.type !== NodeSizeEnum.SMALL) {
                        for (let rc = 0; rc < node.shapeNodes.length; rc++) {
                            for (let cc = 0; cc < node.shapeNodes[rc].shapeNodeData.length; cc++) {
                                if (!node.shapeNodes[rc].isFooterRow) {
                                    const tableCol: ShapeNodeDataModel = node.shapeNodes[rc].shapeNodeData[cc];
                                    if (tableCol.columnText !== '' && tableCol.columnText !== undefined) {
                                        containerNode.addRow(initialLocation, initialLocation,
                                            tableCol.columnText,
                                            tableCol.columnTextColor,
                                            tableCol.columnImageSrcUrl,
                                            tableCol.vo);
                                        count++;
                                    }

                                }
                            }
                        }
                    }
                    nodes[node.nodeID] = containerNode.container;
                    this.incomingLinks = 0;
                    this.outgoingLinks = 0;
                });
                this.createLinks(nodeLinks, nodes);
            }
        }
    }

    private countIncomingOutgoingNodes(tableNodeLinks: NodeLinkModel[], nodeID: number) {
        tableNodeLinks.forEach(link => {
            if (link.startNodeId === nodeID) {
                this.outgoingLinks++;
            } else if (link.endNodeId === nodeID) {
                this.incomingLinks++;
            }
        });
    }

    private createLinks(tableNodeLinks: NodeLinkModel[], nodes: any) {
        if (tableNodeLinks.length > 0) {
            tableNodeLinks.forEach(link => {
                const connectionLink = this.diagram.getFactory().createDiagramLink(nodes[link.startNodeId], nodes[link.endNodeId]);
                // remove ArrowHead from Connection Link
                connectionLink.setShape(MindFusion.Diagramming.LinkShape.Bezier);
                connectionLink.setAutoRoute(true);
                // set parent Node Expandable.
                nodes[link.startNodeId].setExpandable(true);

                this.rearrangeDiagram(this.diagram);
                this.diagram.resizeToFitItems();
            });
        }
    }

    private setOffsetFoldedChildNode(node, offset) {
        const links = node.getOutgoingLinks();
        for (let l = 0; l < links.length; l++) {
            const link = links[l];
            const child = link.getDestination();
            const rect = child.getBounds().clone();
            rect.y -= offset;
            child.setBounds(rect, true, true);
            for (let c = 1; c < link.getControlPoints().length - 1; c++) {
                link.getControlPoints()[c].y -= offset;
            }
            link.updateFromPoints();
            this.setOffsetFoldedChildNode(child, offset);
        }
    }

    private setOffsetUnFoldedChildNode(node, offset) {
        const links = node.getOutgoingLinks();
        for (let l = 0; l < links.length; l++) {
            const link = links[l];
            const child = link.getDestination();
            const rect = child.getBounds().clone();
            rect.y += offset;
            child.setBounds(rect, true, true);
            for (let c = 1; c < link.getControlPoints().length - 1; c++) {
                link.getControlPoints()[c].y += offset;
            }
            link.updateFromPoints();
            this.setOffsetUnFoldedChildNode(child, offset);
        }
    }

    private rearrangeDiagram(diagram: any) {
        // Apply Layout
        if (diagram !== undefined) {
            const layout = new MindFusion.Graphs.TreeLayout();
            layout.keepGroupLayout = true;
            layout.direction = MindFusion.Graphs.LayoutDirection.TopToBottom;
            layout.linkType = MindFusion.Graphs.TreeLayoutLinkType.Cascading;
            layout.levelDistance = 20;
            layout.nodeDistance = 20;
            diagram.arrange(layout);
        }
        //var layout = new MindFusion.Graphs.FractalLayout();
        //layout.root = diagram.nodes[0];
        //diagram.arrange(layout);
        //this.diagram.zoomToRect(this.diagram.getViewport());
    }

}
